unit objtodb.reader;

{$mode ObjFPC}{$H+}
{$H+}
{$modeswitch advancedrecords}

interface

uses
  Classes, SysUtils, rtti, typinfo, db, sqldb, objtodb.types;

Type
  { TRTTIDBReader }

  TRTTIDBReader = record
  private
    Ctx : TRTTIContext;
    Conn : TSQLConnection;
    function CreateQuery(aSQL : String) : TSQLQuery;
    Procedure DBFieldToField(DBData : TDataset; aData: Pointer; Fld: TRTTIField);
    Procedure FieldsToParams(Params : TParams; aData: Pointer; Rec : TRttiRecordType);
    function RecordToWhereClause(Rec: TRttiRecordType): String;
    Function RecordToSQL(Rec: TRttiRecordType) : String;
    procedure ValueToParam(V: TValue; Parm: TParam);
  Public
    function ReadFromDB(aData : Pointer; aTypeInfo: PTypeInfo) : Boolean;
    Generic function ReadFromDB<T>(var aData : T) : Boolean;
    class function create(aConnection : TSQLConnection): TRTTIDBReader; static;
    procedure Free;
  end;

implementation

function TRTTIDBReader.CreateQuery(aSQL : String) : TSQLQuery;

begin
  Result:=TSQLQuery.Create(Conn);
  Result.Transaction:=Conn.Transaction;
  Result.UsePrimaryKeyAsKey:=False;
  Result.UniDirectional:=True;
  Result.SQL.Text:=aSQL;
end;

Procedure TRTTIDBReader.DBFieldToField(DBData : TDataset; aData: Pointer; Fld: TRTTIField);

var
  V : TValue;
  DBField : DBFieldAttribute;
  DBFld : TField;
  FieldName, S : String;
  DT : TDateTime;
  U : UnicodeString;

begin
  FieldName:=Fld.Name;
  DBField:=DBFieldAttribute(Fld.GetAttribute(DBFieldAttribute));
  if Assigned(DBField) then
    FieldName:=DBField.Name;
  DBFld:=DBData.FieldByName(FieldName);
  Case DBFld.DataType of
    ftInteger,
    ftSmallint,
    ftWord : TValue.Make(DBFld.AsInteger,fld.FieldType.Handle,V);
    ftLargeInt : TValue.Make(DBFld.AsLargeInt,fld.FieldType.Handle,V);
    ftString :
      begin
      S:=DBFld.AsString;
      TValue.Make(@S,fld.FieldType.Handle,V);
      end;
    ftWideString:
      begin
      U:=DBFld.AsUnicodeString;
      TValue.Make(@U,fld.FieldType.Handle,V);
      end;
    ftDate,
    ftTime,
    ftDateTime:
      begin
      DT:=DBFld.AsDateTime;
      TValue.Make(@DT,fld.FieldType.Handle,V);
      end;
  else
    Raise Exception.Create('Unknown field type: '+GetEnumName(TypeInfo(TFieldType),Ord(DBFld.DataType)))
  end;
  Fld.SetValue(aData,V);
end;

Procedure TRTTIDBReader.ValueToParam(V : TValue; Parm : TParam);

begin
  Case V.Kind of
    tkInteger : Parm.AsInteger:=V.AsInteger;
    tkInt64 : Parm.AsInteger:=V.AsInt64;
    tkAString,
    tkString : Parm.AsString:=V.AsString;
    tkUString : Parm.AsUnicodeString:=V.AsUnicodeString;
    tkWString : Parm.AsUnicodeString:=V.AsUnicodeString;
    tkFloat :
      if V.TypeInfo=TypeInfo(TDateTime) then
        Parm.AsDateTime:=V.AsDateTime
      else
        Parm.AsFloat:=V.AsDouble;
  else
    Raise Exception.Create('Unsupported type')
  end;
end;


procedure TRTTIDBReader.FieldsToParams(Params: TParams; aData: Pointer; Rec: TRTTIRecordType);

var
  Fld : TRttiField;
  DBField : DBFieldAttribute;
  Parm : TParam;
  V : TValue;
  FieldName : string;

begin
  For Fld in Rec.GetFields do
    begin
    FieldName:=Fld.Name;
    if Assigned(Fld.GetAttribute(DBKeyFieldAttribute)) then
      begin
      DBField:=DBFieldAttribute(Fld.GetAttribute(DBFieldAttribute));
      if Assigned(DBField) then
        FieldName:=DBField.Name;
      Parm:=Params.FindParam(FieldName);
      if Assigned(Parm) then
        begin
        V:=Fld.GetValue(aData);
        ValueToParam(V,Parm);
        end;
      end;
    end;
end;

Function TRTTIDBReader.RecordToWhereClause(Rec: TRttiRecordType) : String;

var
  Fld : TRttiField;
  DBField : DBFieldAttribute;
  FieldName : string;

begin
  Result:='';
  For Fld in Rec.GetFields do
    begin
    FieldName:=Fld.Name;
    if Assigned(Fld.GetAttribute(DBKeyFieldAttribute)) then
      begin
      DBField:=DBFieldAttribute(Fld.GetAttribute(DBFieldAttribute));
      if Assigned(DBField) then
        FieldName:=DBField.Name;
      if Result<>'' then
        Result:=Result+' AND ';
      Result:=Result+'('+ FieldName+' = :'+FieldName+')';
      end;
    end;
end;

Function TRTTIDBReader.RecordToSQL(Rec: TRttiRecordType) : String;

var
  Tbl : TableAttribute;
  Fld : TRttiField;
  DBField : DBFieldAttribute;
  FieldName : String;
  TableName : string;

begin
  Result:='';
  For Fld in Rec.GetFields do
    begin
    FieldName:=Fld.Name;
    DBField:=DBFieldAttribute(Fld.GetAttribute(DBFieldAttribute));
    if Assigned(DBField) then
      FieldName:=DBField.Name;
    if Result<>'' then
      Result:=Result+', ';
    Result:=Result+FieldName;
    end;
  TableName:=Rec.Name;
  Tbl:=TableAttribute(Rec.GetAttribute(TableAttribute));
  If Assigned(Tbl) then
    TableName:=Tbl.Table;
  Result:='SELECT '+Result+' FROM '+TableName;
  Result:=Result+' WHERE '+RecordToWhereClause(Rec);
end;

Function TRTTIDBReader.ReadFromDB(aData : Pointer; aTypeInfo: PTypeInfo) : Boolean;

var
  lRtti : TRttiRecordType;
  Qry : TSQLQuery;
  Fld : TRttiField;
  SQL : String;
begin
  lRtti:=Ctx.GetType(aTypeInfo) as TRttiRecordType;
  SQL:=RecordToSQL(lRtti);
  Qry:=CreateQuery(SQL);
  try
    FieldsToParams(Qry.Params,aData,lRtti);
    Qry.Open;
    Result:=not Qry.IsEmpty;
    if Result then
      For Fld in lRtti.GetFields do
        DBFieldToField(Qry,aData,Fld);
  finally
    Qry.Free;
  end;
end;

Generic function TRTTIDBReader.ReadFromDB<T>(var aData : T) : Boolean;

begin
  Result:=ReadFromDB(@aData,TypeInfo(T));
end;

class function TRTTIDBReader.create(aConnection : TSQLConnection): TRTTIDBReader; static;

begin
  Result.Ctx:=TRttiContext.Create(False);
  Result.Conn:=aConnection;
end;

procedure TRTTIDBReader.Free;

begin
  Ctx.Free;
end;

end.

